#
# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.
import logging
import subprocess
import sys
from pathlib import Path
from typing import Union

from ingest_envoy.collectors import (
    ImageCollector,
    SECICollector,
    collect_image_metadata,
)
from ingest_envoy.ingestion_manifest import IngestionManifestBuilder
from ingest_envoy.interfaces import IngestLauncherIF
from ingest_envoy.schema import AbstractTextFile
from ingest_envoy.utilities import IngestType, VLASSIngestType


def trigger_ingest(real_ingest: bool, staging_dir: str) -> int:
    """
    Run ingest

    :param real_ingest: real ingestion vs. testing flag
    :param staging_dir: staging directory to ingest from
    :return: return code
    """
    if real_ingest:
        ingest_process = subprocess.run(
            ["./ingest", "--json", "-p", staging_dir],
            stdout=sys.stdout,
            stderr=sys.stderr,
        )
        return ingest_process.returncode
    else:
        # For local testing
        return 0


class IngestCalibrationLauncher(IngestLauncherIF):
    """Setup and Launch Calibration Ingestion"""

    def __init__(self, parameters: dict):
        self.logger = logging.getLogger("ingest_envoy")
        self.sci_product_type = "calibration"
        self.parameters = parameters
        self.staging_source_dir = self.parameters["staging_area"] + "/" + self.parameters["workflowDir"]

    def launch_ingestion(self) -> int:
        """
        Prepare and run ingestion script

        :return: Return code of ingestion script process
        """
        self.logger.info("RUNNING CALIBRATION INGESTION!")
        self.prepare_for_ingest()

        self.logger.info("Running ingest!")
        return trigger_ingest(self.parameters["useIngest"], self.staging_source_dir)

    def prepare_for_ingest(self):
        """
        Prepare files for ingestion, creation of collection tars and manifest

        :return:
        """
        self.logger.info("Preparing for ingest...")
        # 1. run collection script to create calibration tarfile
        self.run_collection_script()

        # 2. create ingestion manifest
        self.create_manifest()

    def run_collection_script(self):
        """
        Run calibration collection script, creates collection tars
        ex. caltables, weblogs, pipeline-artifacts, etc

        If this was a multi-version request, also package up the initial version weblog

        :return:
        """
        self.logger.info("Collecting calibration tables for staging...")
        workflow_dir = self.parameters["workflowDir"]
        sdm_id = self.parameters["sdmId"]
        cal_processing_datetime = self.parameters["processingStart"]
        # run script
        if self.parameters["multiVersion"] is False:
            collector = subprocess.run(
                ["./calibration-table-collector.sh", workflow_dir, sdm_id, cal_processing_datetime],
                stdout=sys.stdout,
                stderr=sys.stderr,
            )
        else:
            initial_version_dir = self.parameters["initialVersionDir"]

            collector = subprocess.run(
                [
                    "./calibration-table-collector.sh",
                    workflow_dir,
                    sdm_id,
                    cal_processing_datetime,
                    initial_version_dir,
                ],
                stdout=sys.stdout,
                stderr=sys.stderr,
            )
        if collector.returncode != 0:
            self.logger.error("ERROR: Calibration product collection failed!")
            exit(1)

    def create_manifest(self, additional_file=None):
        """
        Create the calibration ingestion manifest

        :param additional_file: Unneeded for calibration ingestion
        :return:
        """
        self.logger.info("Creating ingestion manifest...")
        spl = self.parameters["spl"]
        telescope = self.parameters["telescope"]

        IngestionManifestBuilder(Path(self.staging_source_dir), self.sci_product_type, spl, telescope).build()


class IngestImageLauncher(IngestLauncherIF):
    """Setup and Launch Image Ingestion"""

    def __init__(self, arg_type: Union[IngestType, VLASSIngestType], parameters: dict):
        self.logger = logging.getLogger("ingest_envoy")
        self.type = arg_type
        self.sci_product_type = arg_type.value
        self.parameters = parameters
        self.staging_source_dir = self.parameters["staging_area"] + "/" + self.parameters["workflowDir"]
        self.collector = (
            ImageCollector(self.parameters) if isinstance(arg_type, IngestType) else SECICollector(self.parameters)
        )

    def launch_ingestion(self) -> int:
        """
        Prepare and run ingestion script

        :return: Return code from ingestion script process
        """
        self.logger.info("RUNNING IMAGE INGESTION!")
        self.prepare_for_ingest()

        self.logger.info("Running ingest!")
        return trigger_ingest(self.parameters["useIngest"], self.staging_source_dir)

    def prepare_for_ingest(self):
        """
        Prepare files for ingestion, creation of collection tars and manifest

        :return:
        """
        self.logger.info("Preparing for ingest...")
        # 1. collect products and auxiliary image metadata
        aux_file = self.run_collector()
        # 2. create manifest
        self.create_manifest(additional_file=aux_file)

    def run_collector(self) -> AbstractTextFile:
        """
        Run ImageCollector

        :return: AbstractTtextFile containing additional metadata needed for manifest creation
        """
        # 1. collect products to staging area
        self.collector.collect_products()
        # 2. create auxiliary metadata file and return
        return collect_image_metadata()

    def create_manifest(self, additional_file: AbstractTextFile = None):
        """
        Create image ingestion manifest

        :param additional_file: image ingestion requires additional metadata
        :return:
        """
        self.logger.info("Creating ingestion manifest...")
        # needs to be the CAL locator!!
        spl = self.parameters["calSpl"]
        telescope = self.parameters["telescope"]
        additional_metadata = additional_file

        IngestionManifestBuilder(
            Path(self.staging_source_dir),
            self.sci_product_type,
            spl,
            telescope,
            additional_metadata,
        ).build()


class IngestObservationLauncher(IngestLauncherIF):
    """Setup and Launch Observation Ingestion"""

    def __init__(self, parameters: dict):
        self.logger = logging.getLogger("ingest_envoy")
        self.sci_product_type = "execution_block"
        self.parameters = parameters
        self.staging_source_dir = self.parameters["staging_area"] + "/" + self.parameters["workflowDir"]

    def launch_ingestion(self) -> int:
        """
        Prepare and run ingestion script

        :return: Return code of ingestion script process
        """
        self.logger.info("RUNNING OBSERVATION INGESTION!")
        self.prepare_for_ingest()

        self.logger.info("Running ingest!")
        return trigger_ingest(self.parameters["useIngest"], self.staging_source_dir)

    def prepare_for_ingest(self):
        """
        Prepare files for ingestion, creation of collection tars and manifest

        :return:
        """
        self.logger.info("Preparing for ingest...")
        # 1. run collection script to create calibration tarfile
        # self.run_collection_script()

        # 2. create ingestion manifest
        self.create_manifest()

    def create_manifest(self, additional_file=None):
        """
        Create the observation ingestion manifest

        :param additional_file: Unneeded for observation ingestion
        :return:
        """
        self.logger.info("Creating ingestion manifest...")
        telescope = self.parameters["telescope"]
        filename = self.parameters["filename"]

        IngestionManifestBuilder(
            Path(self.staging_source_dir), self.sci_product_type, None, telescope, filename=filename
        ).build()
